/*
 * kgraft_patch_bsc1053153
 *
 * Fix for CVE-2017-10661, bsc#1053153
 *
 *  Upstream commit:
 *  1e38da300e1e395a15048b0af1e5305bd91402f6
 *
 *  SLE12-SP2 and -SP3 commit:
 *  Fix arrived via 4.4.67, i.e. 0566c1ce767c697df2d49ff8b445f062de2da1bb
 *
 *  SLE12(-SP1) commit:
 *  b25768cd8675f896a5438e2ce5f976ce291fb6c4
 *
 *  Copyright (c) 2017 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/timekeeping.h>
#include <linux/hrtimer.h>
#include <linux/alarmtimer.h>
#include <linux/timerfd.h>
#include <linux/file.h>
#include <linux/slab.h>
#include <linux/anon_inodes.h>
#include <linux/syscalls.h>
#include "shadow.h"

static const unsigned long kgr_shadow_cancel_lock_id =
	KGR_SHADOW_ID(1053153, 0);

struct timerfd_ctx;

static struct list_head *kgr_cancel_list;
static spinlock_t *kgr_cancel_lock;
static const struct file_operations *kgr_timerfd_fops;
static int (*kgr_timerfd_fget)(int fd, struct fd *p);
static enum hrtimer_restart (*kgr_timerfd_tmrproc)(struct hrtimer *htmr);
static enum alarmtimer_restart (*kgr_timerfd_alarmproc)(struct alarm *alarm,
							ktime_t now);
static ktime_t (*kgr_timerfd_get_remaining)(struct timerfd_ctx *ctx);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "cancel_list", (void *)&kgr_cancel_list },
	{ "cancel_lock", (void *)&kgr_cancel_lock },
	{ "timerfd_fops", (void *)&kgr_timerfd_fops },
	{ "timerfd_fget", (void *)&kgr_timerfd_fget },
	{ "timerfd_tmrproc", (void *)&kgr_timerfd_tmrproc },
	{ "timerfd_alarmproc", (void *)&kgr_timerfd_alarmproc },
	{ "timerfd_get_remaining", (void *)&kgr_timerfd_get_remaining },
};

/* from fs/timerfd.c */
struct timerfd_ctx {
	union {
		struct hrtimer tmr;
		struct alarm alarm;
	} t;
	ktime_t tintv;
	ktime_t moffs;
	wait_queue_head_t wqh;
	u64 ticks;
	int clockid;
	short unsigned expired;
	short unsigned settime_flags;	/* to show in fdinfo */
	struct rcu_head rcu;
	struct list_head clist;
	bool might_cancel;
};


/* optimized */
static bool kgr_timerfd_canceled(struct timerfd_ctx *ctx)
{
	if (!ctx->might_cancel || ctx->moffs.tv64 != KTIME_MAX)
		return false;
	ctx->moffs = ktime_mono_to_real((ktime_t){ .tv64 = 0 });
	return true;
}

/* inlined */
static inline bool kgr_isalarm(struct timerfd_ctx *ctx)
{
	return ctx->clockid == CLOCK_REALTIME_ALARM ||
		ctx->clockid == CLOCK_BOOTTIME_ALARM;
}

/* inlined */
static int kgr_timerfd_setup(struct timerfd_ctx *ctx, int flags,
			 const struct itimerspec *ktmr)
{
	enum hrtimer_mode htmode;
	ktime_t texp;
	int clockid = ctx->clockid;

	htmode = (flags & TFD_TIMER_ABSTIME) ?
		HRTIMER_MODE_ABS: HRTIMER_MODE_REL;

	texp = timespec_to_ktime(ktmr->it_value);
	ctx->expired = 0;
	ctx->ticks = 0;
	ctx->tintv = timespec_to_ktime(ktmr->it_interval);

	if (kgr_isalarm(ctx)) {
		alarm_init(&ctx->t.alarm,
			   ctx->clockid == CLOCK_REALTIME_ALARM ?
			   ALARM_REALTIME : ALARM_BOOTTIME,
			   kgr_timerfd_alarmproc);
	} else {
		hrtimer_init(&ctx->t.tmr, clockid, htmode);
		hrtimer_set_expires(&ctx->t.tmr, texp);
		ctx->t.tmr.function = kgr_timerfd_tmrproc;
	}

	if (texp.tv64 != 0) {
		if (kgr_isalarm(ctx)) {
			if (flags & TFD_TIMER_ABSTIME)
				alarm_start(&ctx->t.alarm, texp);
			else
				alarm_start_relative(&ctx->t.alarm, texp);
		} else {
			hrtimer_start(&ctx->t.tmr, texp, htmode);
		}

		if (kgr_timerfd_canceled(ctx))
			return -ECANCELED;
	}

	ctx->settime_flags = flags & TFD_SETTIME_FLAGS;
	return 0;
}

/* new, formerly timerfd_remove_cancel() */
static void kgr__timerfd_remove_cancel(struct timerfd_ctx *ctx)
{
	if (ctx->might_cancel) {
		ctx->might_cancel = false;
		spin_lock(kgr_cancel_lock);
		list_del_rcu(&ctx->clist);
		spin_unlock(kgr_cancel_lock);
	}
}

/* patched, inlined by do_timerfd_settime() */
static void kgr_timerfd_setup_cancel(struct timerfd_ctx *ctx, int flags)
{
	/*
	 * Fix CVE-2017-10661
	 *  +2 lines
	 */
	spinlock_t *shadow_cancel_lock;
	DEFINE_SPINLOCK(shadow_cancel_lock_init);

	/*
	 * Fix CVE-2017-10661
	 *  -3 lines, +12 lines
	 *
	 * This differs from the upstream patch in that we don't take
	 * the (shadowed) per-timerfd_ctx ->cancel_lock
	 * unconditionally: shadow variables are expensive to retrieve
	 * and if the invariant ->clockid doesn't qualify for the
	 * addition to the cancel_list anyway, we return early as an
	 * optimization.
	 *
	 * Also, in case the timerfd_ctx instance was created before
	 * sys_timerfd_create() has been patched, it might lack a
	 * shadowed ->cancel_lock and we try to attach one. Failures
	 * aren't propagated to the caller though because returning
	 * -ENOMEM from the timerfd_settime syscall would be new
	 * behaviour and applications might not be prepared for it.
	 */
	if (ctx->clockid != CLOCK_REALTIME &&
	    ctx->clockid != CLOCK_REALTIME_ALARM)
		return;

	shadow_cancel_lock =
		klp_shadow_get_or_alloc(ctx, kgr_shadow_cancel_lock_id,
					&shadow_cancel_lock_init,
					sizeof(shadow_cancel_lock_init),
					GFP_KERNEL);
	if (likely(shadow_cancel_lock))
		spin_lock(shadow_cancel_lock);
	if ((flags & TFD_TIMER_ABSTIME) && (flags & TFD_TIMER_CANCEL_ON_SET)) {
		if (!ctx->might_cancel) {
			ctx->might_cancel = true;
			spin_lock(kgr_cancel_lock);
			list_add_rcu(&ctx->clist, kgr_cancel_list);
			spin_unlock(kgr_cancel_lock);
		}
	} else if (ctx->might_cancel) {
		/*
		 * Fix CVE-2017-10661
		 *  No actual change at this point.
		 */
		kgr__timerfd_remove_cancel(ctx);
	}
	/*
	 * Fix CVE-2017-10661
	 *  +2 lines
	 */
	if (likely(shadow_cancel_lock))
		spin_unlock(shadow_cancel_lock);
	return;
}

/*
 * Patched for freeing shadow data.
 */
int kgr_timerfd_release(struct inode *inode, struct file *file)
{
	struct timerfd_ctx *ctx = file->private_data;

	/*
	 * Fix CVE-2017-10661
	 *  No actual change at this point.
	 *
	 * Note that this differs from the upstream commit which calls
	 * the ->cancel_lock taking timerfd_remove_cancel() (rather
	 * than the non-locking __timerfd_remove_cancel()) here.
	 *
	 * The rationale is that a struct file's ->release() can't
	 * race because it's called only when the last reference is
	 * gone.
	 *
	 * In fact, if a race had been possible here, we would have
	 * been in trouble anyway because we'll deallocate the shadow
	 * data in a second and if that could still be in use by
	 * somebody else, then things would explode.
	 */
	kgr__timerfd_remove_cancel(ctx);
	/*
	 * Fix CVE-2017-10661
	 *  +1 line
	 *
	 * Free the shadow data, if any.
	 */
	klp_shadow_free(ctx, kgr_shadow_cancel_lock_id);

	if (kgr_isalarm(ctx))
		alarm_cancel(&ctx->t.alarm);
	else
		hrtimer_cancel(&ctx->t.tmr);
	kfree_rcu(ctx, rcu);
	return 0;
}

/* patched because it inlines timerfd_setup_cancel() */
int kgr_do_timerfd_settime(int ufd, int flags,
		const struct itimerspec *new,
		struct itimerspec *old)
{
	struct fd f;
	struct timerfd_ctx *ctx;
	int ret;

	if ((flags & ~TFD_SETTIME_FLAGS) ||
	    !timespec_valid(&new->it_value) ||
	    !timespec_valid(&new->it_interval))
		return -EINVAL;

	ret = kgr_timerfd_fget(ufd, &f);
	if (ret)
		return ret;
	ctx = f.file->private_data;

	kgr_timerfd_setup_cancel(ctx, flags);

	/*
	 * We need to stop the existing timer before reprogramming
	 * it to the new values.
	 */
	for (;;) {
		spin_lock_irq(&ctx->wqh.lock);

		if (kgr_isalarm(ctx)) {
			if (alarm_try_to_cancel(&ctx->t.alarm) >= 0)
				break;
		} else {
			if (hrtimer_try_to_cancel(&ctx->t.tmr) >= 0)
				break;
		}
		spin_unlock_irq(&ctx->wqh.lock);
		cpu_relax();
	}

	/*
	 * If the timer is expired and it's periodic, we need to advance it
	 * because the caller may want to know the previous expiration time.
	 * We do not update "ticks" and "expired" since the timer will be
	 * re-programmed again in the following timerfd_setup() call.
	 */
	if (ctx->expired && ctx->tintv.tv64) {
		if (kgr_isalarm(ctx))
			alarm_forward_now(&ctx->t.alarm, ctx->tintv);
		else
			hrtimer_forward_now(&ctx->t.tmr, ctx->tintv);
	}

	old->it_value = ktime_to_timespec(kgr_timerfd_get_remaining(ctx));
	old->it_interval = ktime_to_timespec(ctx->tintv);

	/*
	 * Re-program the timer to the new value ...
	 */
	ret = kgr_timerfd_setup(ctx, flags, new);

	spin_unlock_irq(&ctx->wqh.lock);
	fdput(f);
	return ret;
}

/* patched */
__SYSCALL_DEFINEx(2, _kgr_timerfd_create, int, clockid, int, flags)
{
	int ufd;
	struct timerfd_ctx *ctx;
	/*
	 * Fix CVE-2017-10661
	 *  +1 line
	 */
	DEFINE_SPINLOCK(shadow_cancel_lock_init);

	/* Check the TFD_* constants for consistency.  */
	BUILD_BUG_ON(TFD_CLOEXEC != O_CLOEXEC);
	BUILD_BUG_ON(TFD_NONBLOCK != O_NONBLOCK);

	if ((flags & ~TFD_CREATE_FLAGS) ||
	    (clockid != CLOCK_MONOTONIC &&
	     clockid != CLOCK_REALTIME &&
	     clockid != CLOCK_REALTIME_ALARM &&
	     clockid != CLOCK_BOOTTIME &&
	     clockid != CLOCK_BOOTTIME_ALARM))
		return -EINVAL;

	ctx = kzalloc(sizeof(*ctx), GFP_KERNEL);
	if (!ctx)
		return -ENOMEM;

	/*
	 * Fix CVE-2017-10661
	 *  +8 lines
	 */
	if (!klp_shadow_alloc(ctx, kgr_shadow_cancel_lock_id,
			      &shadow_cancel_lock_init,
			      sizeof(shadow_cancel_lock_init),
			      GFP_KERNEL)) {
		kfree(ctx);
		return -ENOMEM;
	}

	init_waitqueue_head(&ctx->wqh);
	ctx->clockid = clockid;

	if (kgr_isalarm(ctx))
		alarm_init(&ctx->t.alarm,
			   ctx->clockid == CLOCK_REALTIME_ALARM ?
			   ALARM_REALTIME : ALARM_BOOTTIME,
			   kgr_timerfd_alarmproc);
	else
		hrtimer_init(&ctx->t.tmr, clockid, HRTIMER_MODE_ABS);

	ctx->moffs = ktime_mono_to_real((ktime_t){ .tv64 = 0 });

	ufd = anon_inode_getfd("[timerfd]", kgr_timerfd_fops, ctx,
			       O_RDWR | (flags & TFD_SHARED_FCNTL_FLAGS));
	/*
	 * Fix CVE-2017-10661
	 *  -2 lines, +4 lines
	 */
	if (ufd < 0) {
		klp_shadow_free(ctx, kgr_shadow_cancel_lock_id);
		kfree(ctx);
	}

	return ufd;
}
/* alias for usedsymbols */
asmlinkage long kgr_sys_timerfd_create(int clockid, int flags)
	__attribute__((alias("SyS_kgr_timerfd_create")));


static int kgr_patch_bsc1053153_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc1053153_init(void)
{
	int ret;

	ret = kgr_shadow_init();
	if (ret)
		return ret;
	return kgr_patch_bsc1053153_kallsyms();
}

void kgr_patch_bsc1053153_cleanup(void)
{
	kgr_shadow_cleanup();
}
